from pathlib import Path
from typing import Union

from appdirs import user_data_dir

DATA_DIR = Path(user_data_dir()) / 'inaturalist'
DEFAULT_DB_PATH = DATA_DIR / 'inaturalist-open-data.db'
DEFAULT_DB_URI = f'sqlite:////{DEFAULT_DB_PATH}'

ARCHIVE_NAME = 'inaturalist-open-data-latest.tar.gz'
BUCKET_NAME = 'inaturalist-open-data'
METADATA_KEY = f'metadata/{ARCHIVE_NAME}'
PHOTO_BASE_URL = 'https://inaturalist-open-data.s3.amazonaws.com/photos/'

PathOrStr = Union[Path, str]



"""Progress bar configuration"""
from typing import Optional, Tuple

from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    SpinnerColumn,
    TaskID,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

ProgressTask = Tuple[Progress, TaskID]


def get_progress(total: int, description: str = 'Loading') -> ProgressTask:
    progress = Progress(
        '[progress.description]{task.description}',
        BarColumn(),
        '[green]{task.completed}/{task.total}',
        '[progress.percentage]{task.percentage:>3.0f}%',
        TimeRemainingColumn(),
    )
    return progress, get_task(progress, total, description)


def get_download_progress(total: int, description: str = 'Downloading') -> ProgressTask:
    progress = Progress(
        '[progress.description]{task.description}',
        BarColumn(),
        '[progress.percentage]{task.percentage:>3.0f}%',
        TransferSpeedColumn(),
        DownloadColumn(),
        TimeRemainingColumn(),
    )
    return progress, get_task(progress, total, description)


def get_spinner_progress(description: str = 'Loading') -> Progress:
    progress = Progress('[progress.description]{task.description}', SpinnerColumn(style='green'))
    get_task(progress, None, description)
    return progress


def get_task(progress, total: Optional[int], description: str) -> TaskID:
    return progress.add_task(f'[cyan]{description}...', total=total)



"""Functions for downloading and extracting inaturalist open data"""
# TODO: Download separate archive files in parallel?
from datetime import datetime, timezone
from io import FileIO
from os.path import basename, getsize
from pathlib import Path
from tarfile import TarFile
from time import time

import boto3
from botocore import UNSIGNED
from botocore.config import Config
from rich import print

from .constants import ARCHIVE_NAME, BUCKET_NAME, DATA_DIR, METADATA_KEY, PathOrStr
from .progress import get_download_progress


class FlatTarFile(TarFile):
    """Extracts all archive contents to a flat base directory, ignoring archive subdirectories"""

    def extract(self, member, path="", **kwargs):
        if member.isfile():
            member.name = basename(member.name)
            super().extract(member, path, **kwargs)


class ProgressIO(FileIO):
    """File object wrapper that updates read progress"""

    def __init__(self, path, *args, **kwargs):
        self._total_size = getsize(path)
        self.progress, self.task = get_download_progress(self._total_size, 'Extracting')
        super().__init__(path, *args, **kwargs)

    def read(self, size):
        self.progress.update(self.task, advance=size)
        return super().read(size)


def download_metadata(download_dir: PathOrStr = DATA_DIR, verbose: int = 0):
    """Download and extract metadata archive. Reuses local data if already exists and is up to date.

    Args:
        download_path: Optional file path to download to
    """
    download_dir = Path(download_dir).expanduser()
    download_dir.mkdir(parents=True, exist_ok=True)
    download_file = download_dir / ARCHIVE_NAME

    # Skip download if we're already up to date
    if not check_download(download_file):
        return

    # Otherwise, download and extract files
    print(f'Downloading to: {download_file}')
    get_s3_file(BUCKET_NAME, METADATA_KEY, download_file)
    progress_file = ProgressIO(download_file)
    with FlatTarFile.open(fileobj=progress_file) as archive, progress_file.progress:
        archive.extractall(path=download_dir)


def check_download(download_file: Path) -> bool:
    """Check if the file already exists locally, and if a newer version is not yet available"""
    new_download = True
    if download_file.exists():
        s3_mtime = get_s3_mtime(BUCKET_NAME, METADATA_KEY)
        file_mtime = get_file_mtime(download_file)
        if s3_mtime > file_mtime:
            print(f'[cyan]File exists, but is out of date:[/cyan] {download_file}')
        else:
            print(f'[cyan]File already exists and is up to date:[/cyan] {download_file}')
            estimate_next_release(s3_mtime)
            new_download = False
    return new_download


def get_file_mtime(file: Path) -> datetime:
    """Get the modified time of a file, as a timezone-aware datetime"""
    file_mtime = file.stat().st_mtime
    return datetime.fromtimestamp(file_mtime).astimezone()


#  TODO: code reuse
def get_s3_mtime(bucket_name: str, key: str) -> datetime:
    """Get the modified time of an S3 file"""
    s3 = boto3.client('s3', config=Config(signature_version=UNSIGNED))
    head = s3.head_object(Bucket=bucket_name, Key=key)
    return head['LastModified']


def get_s3_file(bucket_name: str, key: str, download_file: PathOrStr):
    """Download a file from S3, with progress bar"""
    # Get file size for progress bar
    s3 = boto3.client('s3', config=Config(signature_version=UNSIGNED))
    head = s3.head_object(Bucket=bucket_name, Key=key)
    file_size = head['ContentLength']

    # Download file with a callback to periodically update progress
    progress, task = get_download_progress(file_size)
    with progress:
        s3.download_file(
            Bucket=bucket_name,
            Key=key,
            Filename=str(download_file),
            Callback=lambda n_bytes: progress.update(task, advance=n_bytes),
        )


def estimate_next_release(s3_mtime: datetime):
    """Get an estimate of time until next update. Updates are roughly monthly, but not necessarily
    on the same day each month.
    """
    elapsed = datetime.now(timezone.utc) - s3_mtime
    est_release_days = 31 - elapsed.days
    print(f'[cyan]Possible new release in ~[magenta]{est_release_days} [cyan]days')


if __name__ == '__main__':
    start = time()
    download_metadata()
    print(f'Finished in {time() - start:.2f} seconds')



# flake8: noqa: F401

try:
    from .database import load_all
    from .download import download_metadata
    from .models import *

# Ignore ImportErrors, if this is imported outside a virtualenv at build time
except ImportError as e:
    print(e)



from dataclasses import field

from sqlalchemy import Column
from sqlalchemy.orm import registry

Base = registry()


def sa_field(col_type, index: bool = False, primary_key: bool = False, **kwargs):
    """Get a dataclass field with SQLAlchemy column metadata"""
    column = Column(col_type, index=index, primary_key=primary_key)
    return field(**kwargs, metadata={"sa": column})



from dataclasses import dataclass

from sqlalchemy import ForeignKey, Integer, String
from sqlalchemy.orm import relationship, synonym

from . import Base, sa_field


@Base.mapped
@dataclass
class Photo:
    __tablename__ = 'photo'
    __sa_dataclass_metadata_key__ = 'sa'

    uuid: str = sa_field(String, init=False, primary_key=True, index=True)
    photo_id: int = sa_field(Integer, default=None)
    observation_uuid: str = sa_field(
        ForeignKey('observation.uuid'), default=None, primary_key=True, index=True
    )
    user_id: int = sa_field(ForeignKey('user.id'), default=None, index=True)
    extension: str = sa_field(String, default=None)
    license: str = sa_field(String, default=None)
    width: int = sa_field(Integer, default=None)
    height: int = sa_field(Integer, default=None)
    position: int = sa_field(Integer, default=None)

    observation = relationship('Observation', back_populates='photos')
    user = relationship('User', back_populates='photos')

    # Aliases for columns as provided in the CSV file
    photo_uuid = synonym('observation_uuid')
    observer_id = synonym('user_id')



from dataclasses import dataclass

from sqlalchemy import Integer, String

from .base import Base, sa_field


@Base.mapped
@dataclass
class User:
    __tablename__ = 'user'
    __sa_dataclass_metadata_key__ = 'sa'

    id: int = sa_field(Integer, init=False, primary_key=True, index=True)
    login: str = sa_field(String, default=None, index=True)
    name: str = sa_field(String, default=None)



from logging import basicConfig, getLogger

import click
from click_help_colors import HelpColorsGroup

from .constants import DATA_DIR, DEFAULT_DB_URI
from .database import MODEL_NAMES, create_tables, load_all
from .download import download_metadata

TABLES = ['observation', 'photo', 'taxon', 'user']
VERBOSITY_LOG_LEVELS = {0: 'ERROR', 1: 'WARN', 2: 'INFO', 3: 'DEBUG'}

download_dir_option = click.option(
    '-d', '--download-dir', default=DATA_DIR, help='Alternate path for downloads'
)
init_option = click.option(
    '-i',
    '--init',
    is_flag=True,
    help="Just initialize the database with tables + indexes without loading data",
)
tables_option = click.option(
    '-t',
    '--tables',
    type=click.Choice(MODEL_NAMES.keys(), case_sensitive=False),
    multiple=True,
    help='Load only these specific tables',
)
uri_option = click.option(
    '-u', '--uri', default=DEFAULT_DB_URI, help='Alternate database URI to connect to'
)


@click.group(cls=HelpColorsGroup, help_headers_color='blue', help_options_color='cyan')
@click.option('-v', '--verbose', count=True, help='Show more detailed output')
@click.pass_context
def pynat(ctx, verbose):
    """Commands for working with inaturalist open data"""
    ctx.ensure_object(dict)
    verbose = min(verbose, 3)
    ctx.obj['verbose'] = verbose

    # Configure logging based on verbosity level
    basicConfig(level=VERBOSITY_LOG_LEVELS[verbose])
    if verbose == 1:
        getLogger('pyinaturalist_open_data').setLevel('INFO')


@pynat.command()
@download_dir_option
@uri_option
@click.pass_context
def load(ctx, download_dir, uri):
    """Download and load all data into a database. By default, this will create/update a SQLite
    database. Reuses local data if already exists and is up to date.
    """
    download_metadata(download_dir, verbose=ctx.obj['verbose'])
    load_all(download_dir=download_dir, uri=uri, verbose=ctx.obj['verbose'])


@pynat.command(short_help='Download latest archive, if an update is available')
@download_dir_option
@click.pass_context
def dl(ctx, download_dir):
    """Download and extract all files in the inaturalist open data archive.
    Reuses local data if it exists and is up to date.
    """
    download_metadata(download_dir, verbose=ctx.obj['verbose'])


@pynat.command()
@download_dir_option
@init_option
@tables_option
@uri_option
@click.pass_context
def db(ctx, download_dir, init, tables, uri):
    """Load contents of CSV files into a database. Also creates tables and indexes, if they don't
    already exist.
    """
    if init:
        create_tables(uri=uri)
    else:
        load_all(download_dir=download_dir, tables=tables, uri=uri, verbose=ctx.obj['verbose'])


if __name__ == '__main__':
    pynat()



"""Functions for loading inaturalist open data into a database"""
import csv
from logging import getLogger
from pathlib import Path
from time import time
from typing import List

from dateutil.parser import parse as parse_date
from rich import print
from sqlalchemy import create_engine, delete, insert
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Session
from sqlalchemy.sql.sqltypes import Boolean, DateTime

from .constants import DATA_DIR, DEFAULT_DB_URI, PathOrStr
from .models import Base, Observation, Photo, Taxon, User
from .progress import get_progress

MODEL_CSV_FILES = {
    Observation: 'observations.csv',
    Photo: 'photos.csv',
    Taxon: 'taxa.csv',
    User: 'observers.csv',
}
MODEL_NAMES = {
    'observation': Observation,
    'photo': Photo,
    'taxon': Taxon,
    'user': User,
}

logger = getLogger(__name__)


def create_tables(engine=None, uri: str = DEFAULT_DB_URI):
    """Create tables + indexes based on models, if they don't already exist

    Args:
        engine: A SQLAlchemy engine to connect with
        uri: Alternate database URI to connect to
    """
    print('[cyan]Creating tables')
    engine = engine or create_engine(uri)
    Base.metadata.create_all(engine)
    print('[cyan]Done')


def load_all(
    download_dir: PathOrStr = DATA_DIR,
    tables: List[str] = None,
    uri: str = DEFAULT_DB_URI,
    verbose: int = 0,
):
    """Load contents of CSV files into a database. This is just here for convenience since it works
    with all supported DB dialects. DB-specific bulk inserts are faster, if available.

    Args:
        download_dir: Alternate path for downloads
        tables: Specific table(s) to load instead of all
        uri: Alternate database URI to connect to
        verbose: Show additional output
    """
    session = Session(bind=create_engine(uri, echo=verbose == 3))
    create_tables(session.get_bind())
    print(f'[cyan]Loading data from [magenta]{DATA_DIR}/*.csv[cyan]...')

    if tables:
        models = [MODEL_NAMES[table] for table in tables]
    else:
        models = [User, Taxon, Observation, Photo]

    for model in models:
        try:
            start = time()
            load_table(session, model, Path(download_dir))
            if verbose:
                print(f'Finished in {time() - start:.2f} seconds')
        except OperationalError as e:
            print(e)

    print(f'[cyan]Finished populating database at[/cyan] {uri}')


def load_table(session, model, download_dir: Path):
    """Load CSV contents into a table, with progress bar"""
    csv_path = download_dir / MODEL_CSV_FILES[model]
    progress, task = get_progress(0, f'Loading [magenta]{model.__name__}[/magenta] records')

    # Clear the table and insert everything; faster than checking if individual rows exist
    session.execute(delete(model))
    with progress, open(csv_path) as csv_file:
        num_lines = sum(1 for _ in open(csv_path)) - 1
        progress.update(task, total=num_lines)
        reader = csv.reader(csv_file, delimiter='\t')
        next(reader)  # Skip headers

        for chunk in read_chunks(reader, progress, task):
            session.execute(insert(model), [map_columns(model, row) for row in chunk])
        session.commit()


def read_chunks(reader, progress, task, chunksize=20000):
    """Read a CSV file in chunks"""
    chunk = []
    for i, line in enumerate(reader):
        if i % chunksize == 0 and i > 0:
            yield chunk
            chunk = []
            progress.update(task, advance=chunksize)
        chunk.append(line)
    progress.update(task, advance=i % chunksize + 1)
    yield chunk


def map_columns(model, row):
    """Get a column mapping from model properties to CSV columns. These are already in the same
    order, but since we're using a lower-level bulk insert, this must be explicit.
    """

    return {col.name: convert_types(col, row[i]) for i, col in enumerate(model.__mapper__.columns)}


def convert_types(col, value):
    """Do any type conversions that the database can't do automatically"""
    if isinstance(col.type, DateTime):
        value = parse_date(value) if value else None
    elif isinstance(col.type, Boolean):
        value = str(value).lower() == 'true'
    # Replace empty strings and whitespace with None
    elif not (value and str(value).strip()):
        value = None
    return value


if __name__ == '__main__':
    start = time()
    load_all()
    print(f'Finished in {time() - start:.2f} seconds')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyinaturalist_open_data', 'pyinaturalist_open_data.models']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4,<2.0',
 'boto3>=1.17,<2.0',
 'click-help-colors>=0.9,<0.10',
 'click>=7.1,<7.2',
 'pyinaturalist>=0.12.0,<0.13.0',
 'requests-cache>=0.6.3,<0.7.0',
 'rich>=10.0',
 'sqlalchemy>=1.4,<2.0']

entry_points = \
{'console_scripts': ['pynat = pyinaturalist_open_data.cli:pynat']}

setup_kwargs = {
    'name': 'pyinaturalist-open-data',
    'version': '0.1.1',
    'description': 'Python utilities for working with inaturalist-open-data',
    'long_description': "# pyinaturalist-open-data\n\n[![Build status](https://github.com/JWCook/pyinaturalist-open-data/workflows/Build/badge.svg)](https://github.com/JWCook/pyinaturalist-open-data/actions)\n[![PyPI](https://img.shields.io/pypi/v/pyinaturalist-open-data?color=blue)](https://pypi.org/project/pyinaturalist-open-data)\n[![PyPI - Python Versions](https://img.shields.io/pypi/pyversions/pyinaturalist-open-data)](https://pypi.org/project/pyinaturalist-open-data)\n[![PyPI - Format](https://img.shields.io/pypi/format/pyinaturalist-open-data?color=blue)](https://pypi.org/project/pyinaturalist-open-data)\n\n**This is a work in progress and not yet complete!**\n\n**pyinaturalist-open-data** is a python library and CLI tool for working with\n[inaturalist-open-data](https://github.com/inaturalist/inaturalist-open-data).\nIts goal is to make it easy to import and use this dataset in a python application backed by\n[any SQLAlchemy-compatible database engine](https://docs.sqlalchemy.org/en/14/core/engines.html#supported-databases)\n(SQLite by default), or simply for local data exploration.\n\nSee the CLI in action here or on [asciinema](https://asciinema.org/a/412516):\n[![asciicast](docs/pynat-demo.gif)](https://asciinema.org/a/412516)\n\n## Installation\nInstall with pip:\n```\npip install pyinaturalist-open-data\n```\n\nOr for local development:\n```bash\ngit clone https://github.com/JWCook/pyinaturalist-open-data.git\ncd pyinaturalist-open-data\npip install poetry && poetry install\n```\n\n## Usage\n\nThis package provides the command `pynat`. See `--help` for commands and options:\n```\nUsage: pynat [OPTIONS] COMMAND [ARGS]...\n\n  Commands for working with inaturalist open data\n\nOptions:\n  -v, --verbose  Show more detailed output\n  --help         Show this message and exit.\n\nCommands:\n  db    Load contents of CSV files into a database\n  dl    Download and extract inaturalist open data archive\n  init  Just create tables (if they don't already exist) without populating...\n  load  Download and load all data into a database.\n```\n\n### Run everything\nThe simplest command is `load`, which runs all steps:\n1. Download and extract the dataset\n2. Create database tables and indices\n3. Load the data into the database\n\nOptions:\n```\nUsage: pynat load [OPTIONS]\n\nOptions:\n  -d, --download-dir TEXT  Alternate path for downloads\n  -u, --uri TEXT           Alternate database URI to connect to\n  --help                   Show this message and exit.\n```\n\nBy default, this will create a new SQLite database. Alternatively, you can provide a URI for\n[any supported database](https://docs.sqlalchemy.org/en/14/core/engines.html#supported-databases).\n\n### Run individual steps\nOther commands are available if you only one to run one of those steps at a time.\n\n`dl` command:\n```\nUsage: pynat dl [OPTIONS]\n\n  Download and extract all files in the inaturalist open data archive\n\nOptions:\n  -d, --download-dir TEXT  Alternate path for downloads\n  --help                   Show this message and exit\n```\n\n**Note:** Both `dl` and `load` will reuse local data if already exists and is up to date.\n\n`db` command:\n```\nUsage: pynat db [OPTIONS]\n\n  Load contents of CSV files into a database. Also creates tables and\n  indexes, if they don't already exist.\n\nOptions:\n  -d, --download-dir TEXT         Alternate path for downloads\n  -i, --init                      Just initialize the database with tables\n                                  + indexes without loading data\n  -t, --tables [observation|photo|taxon|user]\n                                  Load only these specific tables\n  -u, --uri TEXT                  Alternate database URI to connect to\n\n  --help                          Show this message and exit.\n```\n\n**Note:** This can take a long time to run. Depending on the database type, you will likely get\nbetter performance with database-specific bulk loading tools (for example, `psql` with [COPY](https://www.postgresql.org/docs/13/sql-copy.html) for PostgreSQL)\n\n### Python package\nTo use as a python package instead of a CLI tool:\n```python\nfrom pyinaturalist_open_data import download_metadata, load_all\n\ndownload_metadata()\nload_all()\n```\n\nFull package documentation on readthedocs will be coming soon.\n\n## Planned features\nSome features I would ideally like to add to this:\n* Performance optimizations\n* Basic querying features\n* Image downloads based on query results\n* Integration with [iNaturalist API](https://api.inaturalist.org/v1/docs/#/) data via\n  [pyinaturalist](https://github.com/JWCook/pyinaturalist-open-data)\n* Integration with CSV data from the [iNaturalist export tool](https://www.inaturalist.org/observations/export)\n",
    'author': 'Jordan Cook',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/JWCook/pyinaturalist-open-data',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import List

from sqlalchemy import Boolean, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property

from .base import Base, sa_field


@Base.mapped
@dataclass
class Taxon:
    __tablename__ = 'taxon'
    __sa_dataclass_metadata_key__ = 'sa'

    id: int = sa_field(Integer, init=False, primary_key=True, index=True)
    ancestry: str = sa_field(String, default=None)
    rank_level: str = sa_field(String, default=None)
    rank: str = sa_field(String, default=None)
    name: str = sa_field(String, default=None, index=True)
    active: bool = sa_field(Boolean, default=None)

    @hybrid_property
    def ancestors(self) -> List[str]:
        return self.ancestry.split('\\')



from dataclasses import dataclass
from datetime import datetime

from sqlalchemy import DateTime, Float, ForeignKey, Integer, String
from sqlalchemy.orm import relationship, synonym

from .base import Base, sa_field


@Base.mapped
@dataclass
class Observation:
    __tablename__ = 'observation'
    __sa_dataclass_metadata_key__ = 'sa'

    uuid: str = sa_field(String, init=False, primary_key=True, index=True)
    user_id: int = sa_field(ForeignKey('user.id'), default=None, index=True)
    latitude: float = sa_field(Float, default=None)
    longitude: float = sa_field(Float, default=None)
    positional_accuracy: int = sa_field(Integer, default=None)
    taxon_id: int = sa_field(ForeignKey('taxon.id'), default=None, index=True)
    quality_grade: str = sa_field(String, default=None, index=True)
    observed_on: datetime = sa_field(DateTime, default=None, index=True)

    user = relationship('User', back_populates='observations')
    taxon = relationship('Taxon', back_populates='observations')

    # Aliases for columns as provided in the CSV file
    observation_uuid = synonym('observation_uuid')
    observer_id = synonym('user_id')
